#!/usr/bin/python
# -*-coding: utf-8-*-

import numpy as np
import scipy.optimize as scopt
import math
import matplotlib.pyplot as plt

# Programm zur Bereitstellung 
# der Funktion f(y,x,t) der
# DAE f((d(x,t))',x,t) = 0
# choose daetype = 'classic' if d(x,t)=x
# choose daetype = 'proper' if ker(f_y) oplus im(d_x) 

def f(y,x,t,daetype,parameter):
    eta = parameter
    if daetype == 'proper':
        f1 = y + x[1]
        f2 = x[0] + eta*t*x[1] - math.exp(-t)
    else: # daetype = 'classic'
        f1 = y[0] + eta*t*y[1] + (1+eta)*x[1]
        f2 = x[0] + eta*t*x[1] - math.exp(-t)
    return np.array([f1,f2])
# end def

# Programm zur Bereitstellung 
# der Funktion d(x,t) der
# DAE f((d(x,t))',x,t) = 0
# choose daetype = 'classic' if d(x,t)=x
# choose daetype = 'proper' if ker(f_y) oplus im(d_x) 
def d(x,t,daetype,parameter):
    eta = parameter
    if daetype == 'proper':
        d = x[0] + eta*t*x[1]
    else:
        d = x
    return d
# end def

# BDF1-Loesung
# ist eine Nullstelle von bdf1_dae_step_fo (nahe (x_k,x_k)).
def bdf1_dae_step_fo(X,xk,tkp1,h,daetype,parameter):
    dX = d(X,tkp1,daetype,parameter) 
    dxk = d(xk,tkp1-h,daetype,parameter) 
    ykp1 = (dX - dxk)/h
    F = f(ykp1,X,tkp1,daetype,parameter)
    return F
# end def

# Programm 9.2
# zur Loesung von DAEs f((d(x,t))',x,t)=0
# BDF1-Verfahren (Ordnung 1)
# input:  Intervallgrenzen t0,tf,
#         Anfangswert x0=x(t0), Schrittweite h
# output: Loesung x(t) an den Punkten t
# benutzte Programme: Nutzerfunktionen f und d
# Aufruf: [x,t] = radauIIA(t0,tf,x0,h);

def bdf1_dae(t0,tf,x0,h,daetype,parameter):
    n = int((tf - t0)/h)
    m = len(x0)

    t = np.zeros(n+1); x = np.zeros((n+1,m))
    t[0] = t0
    x[0,:] = x0
    xk = x0 # starte mit x_k:=x0
    tkp1 = t0 + h
#    tkp1 = t0

    for k in range(n):
        args = (xk,tkp1,h,daetype,parameter)

        # berechne Nullstelle von bdf1_dae_step_fo
        # und nehme x_k als Startnaeherung:
        Xstart = xk
        X = scopt.fsolve(bdf1_dae_step_fo,Xstart,args)
        xkp1 = X # Loesung BDF1-System
        t[k+1] = tkp1
        x[k+1,:] = xkp1   # speichere aktuelle Loesung x_{k+1}
        tkp1 = tkp1 + h       # update t_{k} fuer naechsten Schritt
        xk = xkp1       # update x_{k} fuer naechsten Schritt
    # end for
    return (x,t)
# end def 

def main():
    t0 = 0.0; tf = 4.0
    h = 0.5e-1
    x0 = np.array([1.0,1.0])
#    eta = -0.27
    eta = -0.1
#    eta = 0.1

    x_classic, t_classic = bdf1_dae(t0,tf,x0,h,'classic',eta)
    x_proper, t_proper = bdf1_dae(t0,tf,x0,h,'proper',eta)

    # graphische Darstellung der numerischen Loesung
    fig, axes = plt.subplots(1, 2, figsize=(12.5,4))
    for i in range(2):
        axes[i].plot(t_proper, x_proper[:,i], color="blue", linewidth=2,linestyle="-", label='proper')
        axes[i].plot(t_classic, x_classic[:,i], color="red", linewidth=2,linestyle="-", label='classic')
        axes[i].legend()
        axes[i].set_xlabel('t')
        axes[i].set_ylabel('x'+str(i+1))

    fig.suptitle(f'Loesung des BDF1-Verfahrens für Gear-Petzold-DAE mit $\eta$={eta} und h={h}')
    plt.savefig("BDF1_DAE.png",dpi=300, bbox_inches='tight')
    plt.show()

if __name__ == '__main__':
    main()

