#!/usr/bin/python
# -*-coding: utf-8-*-

import numpy as np
import scipy.optimize as scopt
import math
import matplotlib.pyplot as plt

# Programm zur Bereitstellung 
# der Funktion f(y,x,t) der
# DAE f((d(x,t))',x,t) = 0
# choose daetype = 'classic' if d(x,t)=x
# choose daetype = 'proper' if ker(f_y) oplus im(d_x) 
def f(y,x,t,daetype,parameter):
    eta = parameter
    if daetype == 'proper':
        f1 = y + x[1]
        f2 = x[0] + eta*t*x[1] - math.exp(-t)
    else: # daetype = 'classic'
        f1 = y[0] + eta*t*y[1] + (1+eta)*x[1]
        f2 = x[0] + eta*t*x[1] - math.exp(-t)
    return np.array([f1,f2])
# end def

# Programm zur Bereitstellung 
# der Funktion d(x,t) der
# DAE f((d(x,t))',x,t) = 0
# choose daetype = 'classic' if d(x,t)=x
# choose daetype = 'proper' if ker(f_y) oplus im(d_x) 
def d(x,t,daetype,parameter):
    eta = parameter
    if daetype == 'proper':
        d = x[0] + eta*t*x[1]
    else:
        d = x
    return d
# end def

# Programm 9.2 (radauIIA.py)
# zur Loesung von DAEs f((d(x,t))',x,t)=0
# 2-stufiges Radau-IIA-Verfahren (Ordnung 3)
# input:  Intervallgrenzen t0,tf,
#         Anfangswert x0=x(t0), Schrittweite h
# output: Loesung x(t) an den Punkten t
# benutzte Programme: Nutzerfunktionen f und d
# Aufruf: [x,t] = radauIIA(t0,tf,x0,h);

def radauIIA(t0,tf,x0,h,daetype,parameter):
    n = int((tf - t0)/h)
    m = len(x0)

    t = np.zeros(n+1); x = np.zeros((n+1,m))
    xk = x0 # starte mit x_k:=x0
    tk = t0 # aktueller Zeitpunkt t_k
    t[0] = t0; x[0] = x0

    beta11 = 1.5;  beta12 = 0.5
    beta21 = -4.5; beta22 = 2.5
    c1 = 1.0/3.0; c2 = 1.0

    # Die Stufenloesungen X=(X1,X2) eines Radau-IIA-Schrittes
    # sind eine Nullstelle von radauIIAstep_f (nahe (x_k,x_k)).
    def radauIIAstep_f(X,xk,tk,h,daetype,parameter):
        X1 = X[0:m]
        X2 = X[m:2*m]
        dX1 = d(X1,tk+c1*h,daetype,parameter)
        dX2 = d(X2,tk+c2*h,daetype,parameter)
        dxk = d(xk,tk,daetype,parameter)
        Y1 = (beta11*(dX1-dxk) + beta12*(dX2-dxk))/h
        Y2 = (beta21*(dX1-dxk) + beta22*(dX2-dxk))/h
        F1 = f(Y1,X1,tk+c1*h,daetype,parameter)
        F2 = f(Y2,X2,tk+c2*h,daetype,parameter)
        return np.concatenate((F1,F2))
    # end def 

    for k in range(n):
        args = (xk,tk,h,daetype,parameter)

        # berechne Nullstelle von radauIIAstep_f
        # und nehme (x_k,x_k) als Startnaeherung:
        Xstart = np.concatenate((xk,xk))
        X = scopt.fsolve(radauIIAstep_f,Xstart,args)
        xkp1 = X[m:2*m] # Loesung der letzten Stufe s=2

        tk = tk+h       # update t_{k} fuer naechsten Schritt
        xk = xkp1       # update x_{k} fuer naechsten Schritt
        t[k+1] = tk     # speichere aktuellen Zeitpunkt t_{k+1}
        x[k+1,:] = xk   # speichere aktuelle Loesung x_{k+1}
    # end for
    return (x,t)
# end def 

def main():
    t0 = 0.0; tf = 4.0
    h = 1e-1
    x0 = np.array([1.0,1.0])
#    eta = -0.27
    eta = 0.1

    x_classic, t_classic = radauIIA(t0,tf,x0,h,'classic',eta)
    x_proper, t_proper = radauIIA(t0,tf,x0,h,'proper',eta)

    # graphische Darstellung der numerischen Loesung
    fig, axes = plt.subplots(1, 2, figsize=(12.5,4))
    for i in range(2):
        axes[i].plot(t_proper, x_proper[:,i], color="blue", linewidth=2,linestyle="-", label='proper')
        axes[i].plot(t_classic, x_classic[:,i], color="red", linewidth=2,linestyle="-", label='classic')
        axes[i].legend()
        axes[i].set_xlabel('t')
        axes[i].set_ylabel('x'+str(i+1))

    fig.suptitle(f'Lösung des 2-stufigen Radau-IIA-Verfahrens für Gear-Petzold-DAE mit $\eta$={eta} und h={h}')
    plt.savefig("radauIIA_GP_DAE.png",dpi=300, bbox_inches='tight')
    plt.show()

if __name__ == '__main__':
    main()

