#!/usr/bin/python
# -*-coding: utf-8-*-

import numpy as np
import scipy.optimize as scopt
import math
import matplotlib.pyplot as plt

# Programme zur Bereitstellung der nichtlinearen Funktion f(y,x,t) und d(x,t)
# der DAE f((d(x,t))',x,t) = 0
# input: daetype = 'classic' fuer d(x,t) = x
#        daetype = 'proper', falls ker(f_y) oplus im(d_x) 
#        parameter fuer Parameter der DAE

def f(y,x,t,daetype,parameter):
    C  = parameter[0] # Kapazitaet
    iS = parameter[1] # Sperrsaettigungsstrom
    uT = parameter[2] # Temperaturspannung
    u_input = parameter[3] # Spannung der Spannungsquelle
    i_diode = iS*(math.exp((x[0]-x[1])/uT)-1.0)
    if daetype == 'proper':
        f1 = C*y + i_diode
        f2 = x[2] - i_diode
        f3 = x[1] - u_input
    else:
        f1 = C*y[0] + i_diode
        f2 = x[2] - i_diode
        f3 = x[1] - u_input
    return np.array([f1,f2,f3])
# end def

def d(x,t,daetype,parameter):
    eta = parameter
    if daetype == 'proper':
        d = x[0]
    else:
        d = x
    return d
# end def

# Programm 9.1 (radauIIA_dae_nonlin.py) zur Loesung der DAE f((d(x,t))',x,t)=0
# 2-stufiges Radau-IIA-Verfahren (Ordnung 3)
# input:  Intervallgrenzen t0,tf; Anfangswert x0=x(t0); Schrittweite h
# output: Loesung x(t) an den Punkten t
# benutzte Programme: Nutzerfunktionen f und d
# Aufruf: [x,t] = radauIIA(t0,tf,x0,h)

def radauIIA(t0,tf,x0,h,daetype,parameter):
    n = int((tf - t0)/h)
    m = len(x0)

    t = np.zeros(n+1); x = np.zeros((n+1,m))
    xk = x0 # starte mit x_k:=x0
    tk = t0 # aktueller Zeitpunkt t_k
    t[0] = t0; x[0] = x0

    beta11 = 1.5;  beta12 = 0.5
    beta21 = -4.5; beta22 = 2.5
    a1 = 1.0/3.0; a2 = 1.0

    # Die Stufenloesung X=(X1,X2) eines Radau-IIA-Schrittes
    # ist Nullstelle von radauIIAstep_f (nahe (x_k,x_k)).
    def radauIIAstep_f(X,xk,tk,h,daetype,parameter):
        X1 = X[0:m]
        X2 = X[m:2*m]
        dX1 = d(X1,tk+a1*h,daetype,parameter)
        dX2 = d(X2,tk+a2*h,daetype,parameter)
        dxk = d(xk,tk,daetype,parameter)
        Y1 = (beta11*(dX1-dxk) + beta12*(dX2-dxk))/h
        Y2 = (beta21*(dX1-dxk) + beta22*(dX2-dxk))/h
        F1 = f(Y1,X1,tk+a1*h,daetype,parameter)
        F2 = f(Y2,X2,tk+a2*h,daetype,parameter)
        return np.concatenate((F1,F2))
    # end def 

    for k in range(n):
        args = (xk,tk,h,daetype,parameter)
        # berechne Nullstelle von radauIIAstep_f
        # und nehme (x_k,x_k) als Startnaeherung:
        Xstart = np.concatenate((xk,xk))
        X = scopt.fsolve(radauIIAstep_f,Xstart,args)
        xkp1 = X[m:2*m] # Loesung der letzten Stufe s=2

        tk = tk+h       # update t_k fuer naechsten Schritt
        xk = xkp1       # update x_k fuer naechsten Schritt
        t[k+1] = tk     # speichere aktuellen Zeitpunkt t_{k+1}
        x[k+1,:] = xk   # speichere aktuelle Loesung x_{k+1}
    # end for
    return (x,t)
# end def 

def main():
    t0 = 0.0; tf = 1e-2; h = 1e-5

    C  = 1.5e-14 # Kapazitaet
    iS = 1e-12 # Sperrsaettigungsstrom
    uT = 25e-3 # Temperaturspannung
    u_input = 5.0 # Spannung der Spannungsquelle

    # Startwert x0:
    x0 = np.zeros(3)
    x0[0] = 4.5
    x0[1] = u_input
    x0[2] = iS*(math.exp((x0[0]-x0[1])/uT)-1.0)

    parameter = np.array([C,iS,uT,u_input])

    x, t = radauIIA(t0,tf,x0,h,'classic',parameter)
    
    # graphische Darstellung der numerischen Loesung
    fig, axes = plt.subplots(1, 3, figsize=(18,4))
    for i in range(3):
        axes[i].plot(t, x[:,i], color="blue", linewidth=2,linestyle="-")
        axes[i].set_xlabel('t')
        axes[i].set_ylabel('x'+str(i+1))

    plt.savefig("radauIIA_Schaltungs_DAE.png",dpi=300, bbox_inches='tight')
    plt.show()

if __name__ == '__main__':
    main()

